#!/usr/bin/env python

"""
tree style dictionaries
"""

from toyplot import color


# GLOBALS
COLORS = [color.to_css(i) for i in color.Palette()]
BLACK = color.black



class TreeStyle:
    def __init__(self, tree_style='n'):
        """
        TreeStyle object used to store tree plotting configurations.
        """
        # set default
        self.tree_style = tree_style
        self.edge_type = 'p'
        self.height = None
        self.width = None
        self.orient = 'right'
        self.node_labels = False
        self.node_size = None
        self.node_color = None
        self.vmarker = "o"
        self.node_hover = True
        self.use_edge_lengths = True
        self.tip_labels = True
        self.tip_labels_color = BLACK
        self.tip_labels_align = False
        self.tip_labels_space = None
        self.scalebar = False
        self.padding = 20

        # substyles with default 'normal' styling
        self.edge_style = EdgeStyle()
        self.node_style = NodeStyle()
        self.edge_align_style = EdgeAlignStyle()
        self.node_labels_style = NodeLabelStyle()
        self.tip_labels_style = TipLabelStyle()
        # self.axes_style = AxesStyle()
        #self.admix_edge_style = AdmixEdgeStyle()
        # override defaults with built-in styles


    def __repr__(self):
        "returns an indented dictionary and subdict print"
        view = []
        for (i, j) in self.__dict__.items():
            if not hasattr(j, '_t'):
                view.append("{}: {}".format(i, j))
            else:
                view.append("{}:".format(i))
                for (k, v) in j.__dict__.items():
                    if k != "_t":
                        view.append("    {}: {}".format(k, v))
        return "\n".join(view)


    def __str__(self):
        "returns same as __repr__"
        return self.__repr__()


    # def _update_from_tree_style(self):
    #     "update this styledict with another styledict or a built-in"
    #     # built-in type
    #     if self.tree_style in ("c", "coal"):
    #         self._update_from_dict(coal.__dict__)
    #     elif self.tree_style in ("d", "dark"):
    #         self._update_from_dict(dark.__dict__)
    #     elif self.tree_style in ("m", "multi"):
    #         self._update_from_dict(multi.__dict__)
    #     else:
    #         self._update_from_dict(normal.__dict__)            


    def _update_from_dict(self, kwargs):
        "updates the default attrs as well as those in Style class objects"
        # iterate over each item in kwargs
        for (key, val) in kwargs.items():

            # enter standard items
            if not isinstance(val, dict):
                if key in self.__dict__.keys():
                    if val is not None:
                        self.__setattr__(key, val)

            # extra style are dicts that contain CSS style '-' keys
            else:
                # load the styledict
                styleobj = self.__dict__.get(key)
                alt = {j: i for (i, j) in styleobj._t.items()}
                for (ikey, ival) in val.items():
                    # key may be a mod-key so we'll try to update both since
                    # it doesn't hurt and allows user to use either arg.
                    tkey = styleobj._t.get(ikey)
                    akey = alt.get(ikey)
                    for key in (ikey, tkey, akey):
                        if key in styleobj.__dict__.keys():
                            # update styledict value
                            if ival is not None:
                                styleobj.__setattr__(key, ival)



class EdgeStyle:
    def __init__(self):
        self.stroke = BLACK
        self.stroke_width = 2
        self.stroke_linecap = "round"
        self.opacity = 1
    
        self._t = {
            "stroke": "stroke", 
            "stroke_width": "stroke-width", 
            "stroke_linecap": "stroke-linecap",
            "opacity": "opacity",
        }

    def cssdict(self):
        return {self._t[i]: j for (i, j) in self.__dict__.items() if i != '_t'}

    # a pretty printed styledict
    def __repr__(self):
        view = []
        for (i, j) in self.__dict__.items():
            if i != "_t":
                view.append("{}: {}".format(i, j))
        return "\n".join(view)


class NodeStyle:
    def __init__(self):
        self.fill = COLORS[0]
        self.stroke = "none"
        self.stroke_width = 1
        self._t = {
            "fill": "fill",
            "stroke": "stroke",
            "stroke_width": "stroke-width",
        }

    def cssdict(self):
        return {self._t[i]: j for (i, j) in self.__dict__.items() if i != '_t'}

    # a pretty printed styledict
    def __repr__(self):
        view = []
        for (i, j) in self.__dict__.items():
            if i != "_t":
                view.append("{}: {}".format(i, j))
        return "\n".join(view)


class EdgeAlignStyle:
    def __init__(self):
        self.stroke = "darkgrey"
        self.stroke_width = 2
        self.stroke_linecap = "round"
        self.stroke_dasharray = "2, 4"
        self._t = {
            "stroke": "stroke", 
            "stroke_width": "stroke-width", 
            "stroke_linecap": "stroke-linecap", 
            "stroke_dasharray": "stroke-dasharray"
        }

    def cssdict(self):
        return {self._t[i]: j for (i, j) in self.__dict__.items() if i != '_t'}

    # a pretty printed styledict
    def __repr__(self):
        view = []
        for (i, j) in self.__dict__.items():
            if i != "_t":
                view.append("{}: {}".format(i, j))
        return "\n".join(view)


class NodeLabelStyle:
    def __init__(self):
        self.fill = BLACK
        self.font_size = "10px"
        self._t = {
            "fill": "fill", 
            "font_size": "font-size",
        }

    def cssdict(self):
        return {self._t[i]: j for (i, j) in self.__dict__.items() if i != '_t'}

    # a pretty printed styledict
    def __repr__(self):
        view = []
        for (i, j) in self.__dict__.items():
            if i != "_t":
                view.append("{}: {}".format(i, j))
        return "\n".join(view)


class TipLabelStyle:
    def __init__(self):
        self.fill = BLACK
        self.font_size = "12px"
        self.text_anchor = "start"
        self.text_anchor_shift = None
        self._t = {
            "fill": "fill", 
            "font_size": "font-size", 
            "text_anchor": "text-anchor", 
            "text_anchor_shift": "-toyplot-anchor-shift",
        }

    def cssdict(self):
        return {self._t[i]: j for (i, j) in self.__dict__.items() if i != '_t'}

    # a pretty printed styledict
    def __repr__(self):
        view = []
        for (i, j) in self.__dict__.items():
            if i != "_t":
                view.append("{}: {}".format(i, j))
        return "\n".join(view)

# class AdmixEdgeStyle:
#     def __init__(self):
#         self["admix_edge_style"]["stroke"] = COLORS[0]
#         self["admix_edge_style"]["stroke_linecap"] = "round"
#         self["admix_edge_style"]["stroke_dasharray"] = "2, 4"


# class AxesStyle:
#     def __init__(self):
#         # standard axes names args
#         self.show = False
#         self.padding = 20
#         self.x_show = True
#         self.x_ticks_show = True
#         self.x_ticks_labels_angle = 0
#         self.y_show = True
#         self.y_ticks_show = True
#         self.y_ticks_labels_angle = 0
    
#         # non-standard axes name args
#         self.x_label_color = None
#         self.y_label_color = None        
#         self.y_domain_min = None
#         self.x_domain_min = None
#         self.y_domain_max = None
#         self.x_domain_max = None
#         #self.y.domain.min = ("y.domain.min", None)
#         #self["axes"]["x.domain.min"] = None
#         #self["axes"]["y.domain.max"] = None
#         #self["axes"]["x.domain.max"] = None

#         self._t = {
#             "show": "show",
#             "padding": "padding",
#             "x_show": "x.show",
#             "x_ticks_show": "x.ticks.show",
#             "x_ticks_labels_angle": "x.ticks.labels.angle",
#             "y_show": "y.show",
#             "y_ticks_show": "y.ticks.show",
#             "y_ticks_labels_angle": "y.ticks.labels.angle",
#             "x_label_color": "x_label_color", 
#             "y_label_color": "y_label_color",
#             "x_domain_max": "x.domain.max", 
#             "y_domain_max": "y.domain.max", 
#             "x_domain_min": "x.domain.min", 
#             "y_domain_min": "y.domain.min",            
#         }

#     # a pretty printed styledict
#     def __repr__(self):
#         view = []
#         for (i, j) in self.__dict__.items():
#             if i != '_t':
#                 view.append("{}: {}".format(i, j))
#         return "\n".join(view)


# PREFIX STYLES ---------------------
normal = TreeStyle('normal')
normal.orient = "right"
normal.use_edge_lengths = True
normal.tip_labels_align = False
# normal.axes_style.show = False
# normal.axes_style.padding = 20

simple = TreeStyle("simple")
simple.orient = "right"
simple.use_edge_lengths = False
simple.node_labels = True
simple.node_color = "lightgrey"
simple.node_size = 18
simple.node_style.stroke = "#262626"
simple.node_style.stroke_width = 1


coal = TreeStyle('coal')
coal.edge_type = 'c'
coal.orient = 'down'
coal.use_edge_lengths = True
coal.tip_labels_align = False
coal.node_labels = True
coal.node_size = 15
coal.node_hover = False
coal.tip_labels = False
coal.scalebar = True
# coal.axes_style.show = True
# coal.axes_style.padding = 20
# coal.axes_style.x_show = False
# coal.axes_style.x_ticks_show = False
# coal.axes_style.x_ticks_labels_angle = 0
# coal.axes_style.y_show = True
# coal.axes_style.y_ticks_show = True
# coal.axes_style.y_domain_min = 0


dark = TreeStyle('dark')
dark.orient = 'right'
dark.use_edge_lengths = True
dark.tip_labels_align = True
dark.tip_labels_color = COLORS[3]
dark.edge_style.stroke = COLORS[0]
# dark.axes_style.padding = 20
# dark.axes_style.show = False
# dark.axes_style.x_show = True
# dark.axes_style.x_label_color = COLORS[7]
# dark.axes_style.y_label_color = COLORS[7]


multi = TreeStyle('multi')
multi.edge_type = 'c'
multi.orient = 'down'
multi.use_edge_lengths = True
multi.tip_labels_align = False
multi.node_labels = False
multi.node_size = 0
multi.tip_labels = True
multi.edge_style.opacity = 0.1
multi.tip_labels_style.text_anchor_shift = "15px"
# multi.axes_style.show = False



#     def update_axes(self, sdict):
#         "add scale bar given orientation of tree and internal/external axes"

#         # can we set the domain max past zero but still show ticks only at
#         # domains that are within the tree scale?
#         if sdict["scalebar"]:
#             if self["orient"] in ["right", "left"]:
#                 self["axes"]["show"] = True
#                 self["axes"]["x.show"] = True
#                 self["axes"]["y.show"] = False
#                 self["axes"]["padding"] = 10
#                 self["axes"]["x.ticks.show"] = True
#                 self["axes"]["y.ticks.show"] = False
#                 self["axes"]["y.ticks.labels.angle"] = 0
#                 self["axes"]["x.domain.max"] = 0
#             else:
#                 self["axes"]["show"] = True
#                 self["axes"]["x.show"] = False
#                 self["axes"]["y.show"] = True
#                 self["axes"]["padding"] = 10
#                 self["axes"]["x.ticks.show"] = False
#                 self["axes"]["y.ticks.show"] = True
#                 self["axes"]["y.ticks.labels.angle"] = -90
#                 self["axes"]["y.domain.min"] = 0
#                 #locations = np.linspace()
#                 #self._x_tick_locator = toyplot.locator.Explicit(
#                 #    locations=locations, 
#                 #    format="{:.2f}")

#         if sdict["padding"]:
#             self["axes"]["padding"] = sdict["padding"]


#     def update_canvas(self, sdict):
#         sdict = {i: j for i, j in sdict.items() if j is not None}
#         for key, val in sdict.items():
#             if isinstance(val, dict):
#                 self[key].update(sdict[key])
#             else:
#                 self[key] = val        



#!/usr/bin/env python

from __future__ import print_function, absolute_import

import re
import sys
import requests
import itertools
from decimal import Decimal
from copy import deepcopy

from .ete3mini import TreeNode
from .TreeStyle import TreeStyle, normal, coal, dark, multi, simple
from .Coords import Coords
from .Drawing import Drawing
from .utils import ToytreeError, TreeMod


class Toytree:
    def __init__(self, newick=None, tree_format=0, fixed_order=None):

        # get the tree as a TreeNode object
        self._parse_to_TreeNode(newick, tree_format)

        # set tips order if fixing for multi-tree plotting (default None)
        self._fixed_order = None
        self._fixed_idx = list(range(self.ntips))
        if fixed_order:
            self._set_fixed_order(fixed_order)

        # ladderize the tree unless user fixed order and wants it not.
        if not self._fixed_order:
            self.tree.ladderize()

        # Object for storing default plot settings or saved styles.
        # Calls several update functions when self.draw() to fit canvas.
        self._style = TreeStyle(tree_style='n')

        # Object for plot coordinates. Calls .update() whenever tree modified.
        self._coords = Coords(self)
        self._coords.update()

        # Object for modifying trees beyond root, prune, drop
        self.mod = TreeMod(self)

    # --------------------------------------------------------------------
    # Class definitions 
    # --------------------------------------------------------------------    
    # ... could add __repr__, __iter__, __next__, but .tree has most already
    def __str__(self):
        """ return ascii tree ... (not sure whether to keep this) """
        return self.tree.__str__()

    def __len__(self):
        """ return len of Tree (ntips) """
        return len(self.tree)

    # --------------------------------------------------------------------
    # Loading Newick or ...
    # --------------------------------------------------------------------    
    def _parse_to_TreeNode(self, newick, tree_format):
        """
        Parse the newick string as either text, filepath or URL, and create an
        ete.TreeNode object as .tree. If no newick init an empty TreeNode.
        """
        if newick:
            # is newick a URL or string, path?
            if any((i in newick for i in ("http://", "https://"))):
                response = requests.get(newick)
                response.raise_for_status()
                newick = response.text.strip()
            # create .tree attribute as TreeNode
            self.tree = TreeNode(newick, format=tree_format)
        # otherwise make an empty TreeNode object
        else:
            self.tree = TreeNode()


    def _set_fixed_order(self, fixed_order):
        if fixed_order:
            if set(fixed_order) != set(self.tree.get_leaf_names()):
                raise ToytreeError(
                    "fixed_order must include same tipnames as tree")
            self._fixed_order = fixed_order
            names = self.get_tip_labels()
            self._fixed_idx = [names.index(i) for i in self._fixed_order]

    # --------------------------------------------------------------------
    # properties are not changeable by the user
    # --------------------------------------------------------------------    
    @property
    def features(self):
        return self.tree.features

    @property
    def nnodes(self):
        return sum(1 for i in self.tree.traverse())

    @property
    def ntips(self):
        return sum(1 for i in self.tree.get_leaves())

    @property
    def newick(self, fmt=0):
        "Returns newick fmt=0 represenation of the tree in its current state."
        # checks one of root's children for features and extra feats.
        if self.tree.children:
            features = {"name", "dist", "support", "height", "idx"}
            testnode = self.tree.children[0]
            extrafeat = {i for i in testnode.features if i not in features}
            features.update(extrafeat)
            return self.tree.write(format=fmt)

    # --------------------------------------------------------------------
    # functions to return values from the ete3 .tree object --------------
    # --------------------------------------------------------------------
    def write(self, handle=None, fmt=0):
        if self.tree.children:
            features = {"name", "dist", "support", "height", "idx"}
            testnode = self.tree.children[0]
            extrafeat = {i for i in testnode.features if i not in features}
            features.update(extrafeat)
            if handle:
                self.tree.write(format=fmt, file=handle)
            else:
                return self.tree.write(format=fmt)


    def get_edge_lengths(self):
        """
        Returns edge length values from tree object in node plot order. To
        modify edge length values you must modify nodes in the .tree object
        directly. For example:

        for node in tree.tree.traverse():
            node.dist = 1.
        """
        return self.get_node_values('dist', True, True)


    def get_node_coordinates(self):
        """
        Returns coordinate locations of nodes in the tree as an array. Each
        row is an (x, y) coordinate, ordered by the 'idx' feature of nodes.
        The first ntips rows are the tip coordinates, which can also be 
        returned using .get_tip_coordinates()
        """
        return self._coords.verts


    def get_node_values(
        self, 
        feature=None, 
        show_root=False, 
        show_tips=False, 
        ):
        """
        Returns node values from tree object in node plot order. To modify
        values you must modify the .tree object directly by setting new
        'features'. For example

        for node in tree.tree.traverse():
            node.add_feature("PP", 100)

        By default node and tip values are hidden (set to "") so that they
        are not shown on the tree plot. To include values for these nodes
        use the 'show_root'=True, or 'show_tips'=True arguments.

        tree.get_node_values("support", True, True)

        """
        # access nodes in the order they will be plotted
        ndict = self.get_node_dict(return_internal=True, return_nodes=True)
        nodes = [ndict[i] for i in range(self.nnodes)[::-1]]

        # get features
        if feature:
            vals = [i.__getattribute__(feature) if hasattr(i, feature)
                    else "" for i in nodes]
        else:
            vals = [" " for i in nodes]

        # apply hiding rules
        if not show_root:
            vals = [i if not j.is_root() else "" for i, j in zip(vals, nodes)]
        if not show_tips:
            vals = [i if not j.is_leaf() else "" for i, j in zip(vals, nodes)]

        # convert float to ints for prettier printing unless all floats
        # raise exception and skip if there are true strings (names)
        try:
            if all([Decimal(str(i)) % 1 == 0 for i in vals if i]):
                vals = [int(i) if isinstance(i, float) else i for i in vals]
        except Exception:
            pass

        return vals


    def get_node_dict(self, return_internal=False, return_nodes=False):
        """
        Return node labels as a dictionary mapping {idx: name} where idx is 
        the order of nodes in 'preorder' traversal. Used internally by the
        func .get_node_values() to return values in proper order. 

        return_internal: if True all nodes are returned, if False only tips.
        return_nodes: if True returns TreeNodes, if False return node names.
        """
        if return_internal:
            if return_nodes:
                return {i.idx: i for i in self.tree.traverse("preorder")}
            else:
                return {i.idx: i.name for i in self.tree.traverse("preorder")}
        else:
            if return_nodes:
                return {
                    i.idx: i for i in self.tree.traverse("preorder")
                    if i.is_leaf()
                }
            else:
                return {
                    i.idx: i.name for i in self.tree.traverse("preorder")
                    if i.is_leaf()
                }
               

    def get_tip_coordinates(self, axis=None):
        """
        Returns coordinates of the tip positions for a tree. If no argument
        for axis then a 2-d array is returned. The first column is the x 
        coordinates the second column is the y-coordinates. If you enter an 
        argument for axis then a 1-d array will be returned of just that axis.
        """
        # get coordinates array
        coords = self.get_node_coordinates()
        if axis == 'x':
            return coords[:self.ntips, 0]
        elif axis == 'y':
            return coords[:self.ntips, 1]
        return coords[:self.ntips]


    def get_tip_labels(self):
        """
        Returns tip labels in ladderized order starting from zero axis
        (bottom to top in right-facing trees; left to right in down-facing).
        """
        return self.tree.get_leaf_names()[::-1]


    def copy(self):
        """ returns a deepcopy of the tree object"""
        return deepcopy(self)


    def is_rooted(self):
        """
        Returns False if the tree is unrooted.
        """
        if len(self.tree.children) > 2:
            return False
        return True


    def is_bifurcating(self, include_root=True):
        """
        Returns False if there is a polytomy in the tree, including if the tree
        is unrooted (basal polytomy), unless you use the include_root=False
        argument.
        """
        ctn1 = -1 + (2 * len(self))
        ctn2 = -2 + (2 * len(self))
        if self.is_rooted():
            return bool(ctn1 == sum(1 for i in self.tree.traverse()))
        if include_root:
            return bool(ctn2 == -1 + sum(1 for i in self.tree.traverse()))
        return bool(ctn2 == sum(1 for i in self.tree.traverse()))

    # --------------------------------------------------------------------
    # functions to modify the ete3 tree - MUST CALL ._coords.update()
    # --------------------------------------------------------------------
    def prune(self, node_idx):
        """
        Returns a subtree pruned from the full tree at the selected
        node index. To find the appropriate index try using
        tre.draw(node_labels=True) and use the interactive hover
        feature, or tre.draw(node_labels='idx') to find the 'idx'
        value of the node on which you wish to prune the tree. If you
        simply want to drop tips from the tree rather than prune on an
        internal node, see the .drop_tip() function instead.

        ptre = tre.prune(15)
        """
        # make a deepcopy of the tree
        nself = self.copy()

        # ensure node_idx is int
        node = nself.tree.search_nodes(idx=int(node_idx))[0]
        nself.tree.prune(node)
        nself._coords.update()
        return nself


    def drop_tips(self, tips):
        """
        Returns a copy of the tree with the selected tips pruned from
        the tree. The entered value can be a name or list of names. To
        prune on an internal node to create a subtree of the existing
        tree see the .prune() function instead.

        ptre = tre.drop_tips(['a', 'b'])
        """
        # make a deepcopy of the tree
        nself = self.copy()

        # if tips is a string or Treenode
        if isinstance(tips, str):
            tips = [tips]
        elif isinstance(tips, TreeNode):
            tips = [tips.name]

        keeptips = [i for i in nself.get_tip_labels() if i not in tips]
        nself.tree.prune(keeptips)
        nself._coords.update()
        return nself


    def resolve_polytomy(
        self,
        default_dist=0.0,
        default_support=0.0,
        recursive=False):
        """
        Returns a copy of the tree with resolved polytomies.
        Does not transform tree in-place.
        """
        nself = self.copy()
        nself.tree.resolve_polytomy(
            default_dist=default_dist,
            default_support=default_support,
            recursive=recursive)
        nself._coords.update()        
        return nself


    def rotate_node(self, idx):
        """
        Returns a ToyTree with the selected node rotated for plotting.
        tip colors do not align correct currently if nodes are rotated...
        """
        # make a copy
        revd = {j: i for (i, j) in enumerate(self.get_tip_labels())}
        neworder = {}
        
        # get node at idx
        node = self.tree.search_nodes(idx=int(idx))[0]
        children = node.children
        aa = [[j.name for j in i.get_leaves()] for i in children]
        bb = [[revd[i] for i in j] for j in aa]
        move = max((len(i) for i in bb))

        # newdict
        tdict = {i: None for i in itertools.chain(*aa)}
        cycle = itertools.cycle(itertools.chain(*bb))
        for m in range(move):
            next(cycle)
        for t in tdict:
            tdict[t] = next(cycle)
            
        for key in revd:
            if key in tdict:
                neworder[key] = tdict[key]
            else:
                neworder[key] = revd[key]
        
        revd = {j: i for (i, j) in neworder.items()}
        neworder = [revd[i] for i in range(self.ntips)]
        return Toytree(self.newick, fixed_order=neworder)


    def unroot(self):
        """
        Returns a copy of the tree unrooted. Does not transform tree in-place.
        """
        nself = self.copy()
        nself.tree.unroot()
        nself.tree.ladderize()
        nself._coords.update()
        return nself


    def root(self, outgroup=None, wildcard=None, regex=None):
        """
        Re-root a tree on a selected tip or group of tip names. Returns a 
        copy of the tree, the original tree object is not modified. 

        The new root can be selected by entering either a list of outgroup
        names, by entering a wildcard selector that matches their names, or
        using a regex command to match their names. For example, to root a tree
        on a clade that includes the samples "1-A" and "1-B" you can do any of
        the following:

        rtre = tre.root(outgroup=["1-A", "1-B"])
        rtre = tre.root(wildcard="1-")
        rtre = tre.root(regex="1-[A,B]")
        """
        # make a deepcopy of the tree
        nself = self.copy()

        # get names of outgroup/s using list, wildcard or regex
        og = outgroup
        if og:
            if isinstance(og, str):
                og = [og]
            notfound = [i for i in og if i not in nself.tree.get_leaf_names()]
            if any(notfound):
                raise Exception(
                    "Sample {} is not in the tree".format(notfound))
            outs = [i for i in nself.tree.get_leaf_names() if i in outgroup]

        elif regex:
            outs = [i.name for i in nself.tree.get_leaves()
                    if re.match(regex, i.name)]
            if not any(outs):
                raise Exception("No Samples matched the regular expression")

        elif wildcard:
            outs = [i.name for i in nself.tree.get_leaves()
                    if wildcard in i.name]
            if not any(outs):
                raise Exception("No Samples matched the wildcard")

        else:
            raise Exception(
                "must enter an outgroup, wildcard selector, or regex pattern")

        # get node to use for outgroup
        if len(outs) > 1:
            # check if they're monophyletic
            mbool, mtype, mnames = nself.tree.check_monophyly(
                outs, "name", ignore_missing=True)
            if not mbool:
                if mtype == "paraphyletic":
                    outs = [i.name for i in mnames]
                else:
                    raise Exception(
                        "Tips entered to root() cannot be paraphyletic")
            out = nself.tree.get_common_ancestor(outs)
        else:
            out = outs[0]

        # split root node if more than di- as this is the unrooted state
        if not nself.is_bifurcating():
            nself.tree.resolve_polytomy()

        # root the object with ete's translate
        nself.tree.set_outgroup(out)
        nself._coords.update()

        # get features
        testnode = nself.tree.get_leaves()[0]
        features = {"name", "dist", "support", "height"}
        extrafeat = {i for i in testnode.features if i not in features}
        features.update(extrafeat)

        # if there is a new node now, clean up its features
        nnode = [i for i in nself.tree.traverse() if not hasattr(i, "idx")]
        if nnode:
            # nnode is the node that was added
            # rnode is the location where it *should* have been added
            nnode = nnode[0]
            rnode = [i for i in nself.tree.children if i != out][0]

            # get idxs of existing nodes
            idxs = [int(i.idx) for i in nself.tree.traverse()
                    if hasattr(i, "idx")]

            # newnode is a tip
            if len(outs) == 1:
                nnode.name = str("rerooted")
                rnode.name = out
                rnode.add_feature("idx", max(idxs) + 1)
                rnode.dist *= 2
                sister = rnode.get_sisters()[0]
                sister.dist *= 2
                rnode.support = 100
                for feature in extrafeat:
                    nnode.add_feature(feature, getattr(rnode, feature))
                    rnode.del_feature(feature)

            # newnode is internal
            else:
                nnode.add_feature("idx", max(idxs) + 1)
                nnode.name = str("rerooted")
                nnode.dist *= 2
                sister = nnode.get_sisters()[0]
                sister.dist *= 2
                nnode.support = 100

        # store tree back into newick and reinit Toytree with new newick
        # if NHX format then preserve the NHX features.
        nself.tree.ladderize()
        nself._coords.update()
        return nself        

    # --------------------------------------------------------------------
    # Draw functions imported, but docstring here
    # --------------------------------------------------------------------
    def draw(
        self,
        tree_style=None,
        height=None,
        width=None,
        axes=None,        
        orient=None,
        tip_labels=None,
        tip_labels_color=None,
        tip_labels_style=None,
        tip_labels_align=None,
        tip_labels_space=None,
        node_labels=None,
        node_labels_style=None,
        node_size=None,
        node_color=None,
        node_style=None,
        node_hover=None,
        edge_type=None,
        edge_style=None,
        edge_align_style=None,
        use_edge_lengths=None,
        scalebar=None,
        padding=None,
        **kwargs):
        """
        Plot a Toytree tree, returns a tuple of Toyplot (Canvas, Axes) objects.

        Parameters:
        -----------
        tree_style: str
            One of several preset styles for tree plotting. The default is 'n'
            (normal). Other options inlude 'c' (coalescent), 'd' (dark), and
            'm' (multitree). You also create your own TreeStyle objects.
            The tree_style sets a default set of styling on top of which other
            arguments passed to draw() will override when plotting.

        height: int (optional; default=None)
            If None the plot height is autosized. If 'axes' arg is used then 
            tree is drawn on an existing Canvas, Axes and this arg is ignored.

        width: int (optional; default=None)
            Similar to height (above). 

        axes: Toyplot.Cartesian (default=None)
            A toyplot cartesian axes object. If provided tree is drawn on it.
            If not provided then a new Canvas and Cartesian axes are created
            and returned with the tree plot added to it.

        use_edge_lengths: bool (default=False)
            Use edge lengths from .tree (.get_edge_lengths) else
            edges are set to length >=1 to make tree ultrametric.

        tip_labels: [True, False, list]
            If True then the tip labels from .tree are added to the plot.
            If False no tip labels are added. If a list of tip labels
            is provided it must be the same length as .get_tip_labels().

        tip_labels_color:
            ...

        tip_labels_style:
            ...

        tip_labels_align:
            ...

        node_labels: [True, False, list]
            If True then nodes are shown, if False then nodes are suppressed
            If a list of node labels is provided it must be the same length
            and order as nodes in .get_node_values(). Node labels can be 
            generated in the proper order using the the .get_node_labels() 
            function from a Toytree tree to draw info from the tree features.
            For example: node_labels=tree.get_node_labels("support").

        node_size: [int, list, None]
            If None then nodes are not shown, otherwise, if node_labels
            then node_size can be modified. If a list of node sizes is
            provided it must be the same length and order as nodes in
            .get_node_dict().

        node_color: [list]
            Use this argument only if you wish to set different colors for
            different nodes, in which case you must enter a list of colors
            as string names or HEX values the length and order of nodes in
            .get_node_dict(). If all nodes will be the same color then use
            instead the node_style dictionary:
            e.g., node_style={"fill": 'red'}

        node_style: [dict]

        ...

        node_hover: [True, False, list, dict]
            Default is True in which case node hover will show the node
            values. If False then no hover is shown. If a list or dict
            is provided (which should be in node order) then the values
            will be shown in order. If a dict then labels can be provided
            as well.
        """
        # allow ts as a shorthand for tree_style
        if kwargs.get("ts"):
            tree_style = kwargs.get("ts")

        # start from a fixed tree_style (change this to a function or dict)
        if tree_style in ('c', 'coal'):
            self._style = deepcopy(coal)
        elif tree_style in ('d', 'dark'):
            self._style = deepcopy(dark)
        elif tree_style in ('m', 'multi'):            
            self._style = deepcopy(multi)
        elif tree_style in ('s', 'simple'):
            self._style = deepcopy(simple)
        else:
            self._style = deepcopy(normal)

        # store entered args
        userargs = {
            "height": height,
            "width": width,
            "orient": orient,
            "tip_labels": tip_labels,
            "tip_labels_color": tip_labels_color,
            "tip_labels_style": tip_labels_style,
            "tip_labels_align": tip_labels_align,
            "tip_labels_space": tip_labels_space,
            "node_labels": node_labels,
            "node_labels_style": node_labels_style,
            "node_size": node_size,
            "node_color": node_color,
            "node_style": node_style,
            "node_hover": node_hover,
            "edge_type": edge_type,  
            "edge_style": edge_style,
            "edge_align_style": edge_align_style,
            "use_edge_lengths": use_edge_lengths,
            "scalebar": scalebar, 
            "padding": padding,
        } 

        # update tree_style to custom style with user entered args
        self._style._update_from_dict(
            {i: j for (i, j) in userargs.items() if j is not None})

        # Init Drawing class object.
        draw = Drawing(self)

        # Debug returns the object to test with.
        if kwargs.get("debug"):
            return draw

        # Make plot. If user provided explicit axes then include them.
        canvas, axes = draw.update(axes=axes)
        return canvas, axes



#!/usr/bin/env python

"""Functions to render HTML markup."""

# pylint: disable=function-redefined

from __future__ import division, absolute_import

import base64
import collections
import copy
import functools
import io
import itertools
import json
import string
import uuid
import xml.etree.ElementTree as xml

from multipledispatch import dispatch
import numpy

import toyplot.coordinates
import toyplot.canvas
import toyplot.color
import toyplot.compatibility
import toyplot.mark
import toyplot.marker

from toyplot.html import RenderContext, _color_fixup


_namespace = dict()

#: Decorator for registering custom rendering code.
#:
#: This is only of use when creating your own custom Toyplot marks.  It is
#: not intended for end-users.
#:
#: Example
#: -------
#: To register your own rendering function::
#:
#:     @toyplot.html.dispatch(toyplot.coordinates.Cartesian, MyCustomMark, toyplot.html.RenderContext)
#:     def _render(axes, mark, context):
#:         # Rendering implementation here
dispatch = functools.partial(dispatch, namespace=_namespace)


# ## get canvas instance
# canvas = toyplot.require.instance(c, toyplot.canvas.Canvas)
# canvas.autorender(False)
# # make top-level HTML element
# root_xml = xml.Element(
#     'div',
#     attrib={"class": "toyplot"},
#     id="t" + uuid.uuid4().hex,
#     )
# # Register a Javascript module to keep track of the root id
# context = RenderContext(root=root_xml)
# context.define("toyplot/root/id", value=root_xml.get("id"))
# context.define(
#     "toyplot/root",
#     ["toyplot/root/id"],
#     factory="""function(root_id){return document.querySelector("#" + root_id)};""")
# mark = M
# axes = a


# toytree Marker object which has a custom render function below.
# The init function for creating a Toytree Mark is similar to what
# used to be in the draw() function.
class Toytree(toyplot.mark.Mark):
    def __init__(self, newick):
        super(Toytree, self).__init__()

        self.toytree = toytree.tree(newick)
        self.markers = []
        edge_marker = get_edge_mark(self.toytree)
        text_marker = get_text_mark(self.toytree)
        node_marker = get_node_mark(self.toytree)
        etip_marker = get_etip_mark(self.toytree)
        for marker in [edge_marker, text_marker, node_marker, etip_marker]:
            if marker:
                self.markers.append(marker)

    # domain is the tree/edge domain
    def domain(self, axis):
        return self.markers[0].domain(axis)
        # if axis == "x":
        #   ## max should fit tip names
        #       return toyplot.data.minimax(self._toytree.verts[:, 0])
        # else:
        #   return toyplot.data.minimax(self._toytree.verts[:, 1])

    # Entents depends on tree orientation, since we extend for text direction
    def extents(self, axes):
        extents = toyplot.text.extents(
            self._table[self._text[0]],
            self._table[self._angle[0]],
            self._style)
        return coordinates, extents



# this is a hybrid of Toyplot's Graph, Text, and Scatterplot _render
# functions with just the elements I want selected out. It also
# makes minifying calls to simplify style dictionaries for shared styles
@dispatch(toyplot.coordinates.Cartesian, Toytree, RenderContext)
def _render(axes, mark, context):

    dimension1 = numpy.ma.column_stack(
        [mark._table[key] for key in mark._coordinates[0::2]])
    dimension2 = numpy.ma.column_stack(
        [mark._table[key] for key in mark._coordinates[1::2]])
    if mark._coordinate_axes[0] == "x":
        X = axes.project("x", dimension1)
        Y = axes.project("y", dimension2)
    elif mark._coordinate_axes[0] == "y":
        X = axes.project("x", dimension2)
        Y = axes.project("y", dimension1)

    # group styles with common values
    shared_styles, unique_styles = split_styles(mark)

    # create root Scatterplot element
    mark_xml = xml.SubElement(
        root_xml,
        "g",
        id=context.get_id(mark),
        attrib={"class": "toyplot-mark-Scatterplot"},
    )
    print xml.tostring(mark_xml)

    ##
    mvectors = [
        X.T,
        Y.T,
        [mark._table[key] for key in mark._marker],
        [mark._table[key] for key in mark._mtitle],
    ]

    for x, y, marker, mtitle in zip(*mvectors):
        not_null = numpy.invert(numpy.logical_or(
            numpy.ma.getmaskarray(x), numpy.ma.getmaskarray(y)))

        # create a Series element to hold all data points
        series_xml = xml.SubElement(
            mark_xml,
            "g",
            attrib={
                "class": "toyplot-Series",
                "style": shared_styles,
            },
        )
        print xml.tostring(series_xml)
        print ""

        # subselect markers for missing data
        vals = [
            itertools.count(step=1),
            x[not_null],
            y[not_null],
            marker[not_null],
            mtitle[not_null],
        ]
        for idx, dx, dy, dmarker, dtitle in zip(*vals):
            marker = toyplot.marker.create(
                size=dmarker.size,
                shape=dmarker.shape,
                mstyle=unique_styles["node"][idx],
                lstyle={},
                label=dmarker.label)
            marker_xml = _draw_marker(
                root=series_xml,
                marker=marker,
                cx=dx,
                cy=dy,
                extra_class="toyplot-Datum",
                title=dtitle,
            )
            # print xml.tostring(marker_xml)
            # print ""
            if marker.label:
                _draw_text(
                    root=marker_xml,
                    text=marker.label,
                    style=unique_styles['text'][idx],
                )
            print xml.tostring(marker_xml)
            print ""
            # create new simplified marker... working off Line 3177 in html.py


## tip names text
def get_text_mark(ttree):
    """ makes a simple Text Mark object"""
    
    if ttree._orient in ["right"]:
        angle = 0.
        ypos = ttree.verts[-1*len(ttree.tree):, 1]
        if ttree._kwargs["tip_labels_align"]:
            xpos = [ttree.verts[:, 0].max()] * len(ttree.tree)
            start = xpos
            finish = ttree.verts[-1*len(ttree.tree):, 0]
            align_edges = np.array([(i, i+len(xpos)) for i in range(len(xpos))])
            align_verts = np.array(zip(start, ypos) + zip(finish, ypos))
        else:
            xpos = ttree.verts[-1*len(ttree.tree):, 0]
            
    elif ttree._orient in ['down']:
        angle = -90.
        xpos = ttree.verts[-1*len(ttree.tree):, 0]
        if ttree._kwargs["tip_labels_align"]:
            ypos = [ttree.verts[:, 1].min()] * len(ttree.tree)
            start = ypos
            finish = ttree.verts[-1*len(ttree.tree):, 1]
            align_edges = np.array([(i, i+len(ypos)) for i in range(len(ypos))])
            align_verts = np.array(zip(xpos, start) + zip(xpos, finish))
        else:
            ypos = ttree.verts[-1*len(ttree.tree):, 1]
    
    table = toyplot.data.Table()
    table['x'] = toyplot.require.scalar_vector(xpos)
    table['y'] = toyplot.require.scalar_vector(ypos, table.shape[0])
    table['text'] = toyplot.broadcast.pyobject(ttree.get_tip_labels(), table.shape[0])
    table["angle"] = toyplot.broadcast.scalar(angle, table.shape[0])
    table["opacity"] = toyplot.broadcast.scalar(1.0, table.shape[0])
    table["title"] = toyplot.broadcast.pyobject(None, table.shape[0])
    style = toyplot.style.require(ttree._kwargs["tip_labels_style"],
                                  allowed=toyplot.style.allowed.text)
    default_color = [toyplot.color.black]
    color = toyplot.color.broadcast(
        colors=ttree._kwargs["tip_labels_color"],
        shape=(table.shape[0], 1),
        default=default_color,
        )
    table["fill"] = color[:, 0]
    
    text_mark = toyplot.mark.Text(
        coordinate_axes=['x', 'y'],
        table=table,
        coordinates=['x', 'y'],
        text=["text"],
        angle=["angle"],
        fill=["fill"],
        opacity=["opacity"],
        title=["title"],
        style=style,
        annotation=True,
        filename=None
        )
    return text_mark


## toytree uses edges, no need for vertices
def get_edge_mark(ttree):
    """ makes a simple Graph Mark object"""
    
    ## tree style
    if ttree._kwargs["tree_style"] in ["c", "cladogram"]:
        a=ttree.edges
        vcoordinates=ttree.verts
    else:
        a=ttree._lines               
        vcoordinates=ttree._coords    
   
    ## fixed args
    along='x'
    vmarker='o'
    vcolor=None
    vlshow=False            
    vsize=0.         
    estyle=ttree._kwargs["edge_style"]

    ## get axes
    layout = toyplot.layout.graph(a, vcoordinates=vcoordinates)
    along = toyplot.require.value_in(along, ["x", "y"])
    if along == "x":
        coordinate_axes = ["x", "y"]
    elif along == "y":
        coordinate_axes = ["y", "x"]
        
    ## broadcast args along axes
    vlabel = layout.vids
    vmarker = toyplot.broadcast.pyobject(vmarker, layout.vcount)
    vsize = toyplot.broadcast.scalar(vsize, layout.vcount)
    estyle = toyplot.style.require(estyle, allowed=toyplot.style.allowed.line)

    ## fixed args
    vcolor = toyplot.color.broadcast(colors=None, shape=layout.vcount, default=toyplot.color.black)
    vopacity = toyplot.broadcast.scalar(1.0, layout.vcount)
    vtitle = toyplot.broadcast.pyobject(None, layout.vcount)
    vstyle = None
    vlstyle = None
    
    ## this could be modified in the future to allow diff color edges
    ecolor = toyplot.color.broadcast(colors=None, shape=layout.ecount, default=toyplot.color.black)
    ewidth = toyplot.broadcast.scalar(1.0, layout.ecount)
    eopacity = toyplot.broadcast.scalar(1.0, layout.ecount)
    hmarker = toyplot.broadcast.pyobject(None, layout.ecount)
    mmarker = toyplot.broadcast.pyobject(None, layout.ecount)
    mposition = toyplot.broadcast.scalar(0.5, layout.ecount)
    tmarker = toyplot.broadcast.pyobject(None, layout.ecount)
    
    ## tables are required if I don't want to edit the class
    vtable = toyplot.data.Table()
    vtable["id"] = layout.vids
    for axis, coordinates in zip(coordinate_axes, layout.vcoordinates.T):
        vtable[axis] = coordinates
        #_mark_exportable(vtable, axis)
    vtable["label"] = vlabel
    vtable["marker"] = vmarker
    vtable["size"] = vsize
    vtable["color"] = vcolor
    vtable["opacity"] = vopacity
    vtable["title"] = vtitle

    etable = toyplot.data.Table()
    etable["source"] = layout.edges.T[0]
    #_mark_exportable(etable, "source")
    etable["target"] = layout.edges.T[1]
    #_mark_exportable(etable, "target")
    etable["shape"] = layout.eshapes
    etable["color"] = ecolor
    etable["width"] = ewidth
    etable["opacity"] = eopacity
    etable["hmarker"] = hmarker
    etable["mmarker"] = mmarker
    etable["mposition"] = mposition
    etable["tmarker"] = tmarker
    
    edge_mark = toyplot.mark.Graph(
        coordinate_axes=['x', 'y'],
        ecolor=["color"],
        ecoordinates=layout.ecoordinates,
        efilename=None,
        eopacity=["opacity"],
        eshape=["shape"],
        esource=["source"],
        estyle=estyle,
        etable=etable,
        etarget=["target"],
        ewidth=["width"],
        hmarker=["hmarker"],
        mmarker=["mmarker"],
        mposition=["mposition"],
        tmarker=["tmarker"],
        vcolor=["color"],
        vcoordinates=['x', 'y'],
        vfilename=None,
        vid=["id"],
        vlabel=["label"],
        vlshow=False,
        vlstyle=None,
        vmarker=["marker"],
        vopacity=["opacity"],
        vsize=["size"],
        vstyle=None,
        vtable=vtable,
        vtitle=["title"],
        )
    return edge_mark


## nodes on tree
def get_node_mark(ttree):
    return


## Currently this clobbers some text styling (e.g., alignment-baseline)
## during the 'toyplot.text.layout' restyling. Not sure why.
def split_styles(mark):
    """ get shared styles """
    
    markers = [mark._table[key] for key in mark._marker][0]
    nstyles = []
    for m in markers:
        ## fill and stroke are already rgb() since already in markers
        msty = toyplot.style.combine({
            "fill": m.mstyle['fill'],
            "stroke": m.mstyle['stroke'],
            "opacity": m.mstyle["fill-opacity"],
        }, m.mstyle)
        msty = _color_fixup(msty)
        nstyles.append(msty)
    
    ## uses 'marker.size' so we need to loop over it
    lstyles = []
    for m in markers:
        lsty = toyplot.style.combine({
        "font-family": "Helvetica",
        "-toyplot-vertical-align": "middle",
        "fill": toyplot.color.black,
        "font-size": "%rpx" % (m.size * 0.75),
        "stroke": "none",
        "text-anchor": "middle",
        }, m.lstyle)
        ## update fonts
        fonts = toyplot.font.ReportlabLibrary()
        layout = toyplot.text.layout(m.label, lsty, fonts)
        lsty = _color_fixup(layout.style)
        lstyles.append(lsty)
    
    nallkeys = set(itertools.chain(*[i.keys() for i in nstyles]))
    lallkeys = set(itertools.chain(*[i.keys() for i in lstyles]))
    nuniquekeys = []
    nsharedkeys = []
    for key in nallkeys:
        vals = [nstyles[i].get(key) for i in range(len(nstyles))]
        if len(set(vals)) > 1:
            nuniquekeys.append(key)
        else:
            nsharedkeys.append(key)
    luniquekeys = []
    lsharedkeys = []
    for key in lallkeys:
        vals = [lstyles[i].get(key) for i in range(len(lstyles))]
        if len(set(vals)) > 1:
            luniquekeys.append(key)
        else:
            lsharedkeys.append(key)

    ## keys shared between mark and text markers
    repeated = set(lsharedkeys).intersection(set(nsharedkeys))
    for repeat in repeated:
        ## if same then keep only one copy of it
        lidx = lsharedkeys.index(repeat)
        nidx = nsharedkeys.index(repeat)
        if lsharedkeys[lidx] == nsharedkeys[nidx]:
            lsharedkeys.remove(repeat)
        else:
            lsharedkeys.remove(repeat)
            luniquekeys.append(repeat)
            nsharedkeys.remove(repeat)
            nuniquekeys.append(repeat)
            
    ## check node values
    natt = ["%s:%s" % (key, nstyles[0][key]) for key in sorted(nsharedkeys)]
    latt = ["%s:%s" % (key, lstyles[0][key]) for key in sorted(lsharedkeys)]
    shared_styles = ";".join(natt+latt)
    unique_styles = {
        "node": [{k:v for k,v in nstyles[idx].items() if k in nuniquekeys} for idx in range(len(markers))],
        "text": [{k:v for k,v in lstyles[idx].items() if k in luniquekeys} for idx in range(len(markers))]
    }
    
    return shared_styles, unique_styles



## Unlike toyplot, I do not want to add text styling at draw time, 
## but rather before hand so that we can simplify the HTML. 
## MAYBE, better way to do this by editing HTML after its written
## inside my own render function, so that less breaking with 
## toyplot is needed. This is here for now...
def _draw_text(
    root,
    text,
    x=0,
    y=0,
    style=None,
    angle=None,
    title=None,
    attributes=None,
    layout=None,
    ):

    if not text:
        return

    if attributes is None:
        attributes = {}

    transform = ""
    if x or y:
        transform += "translate(%r,%r)" % (x, y)
    if angle:
        transform += "rotate(%r)" % (-angle)

    group = xml.SubElement(
        root,
        "g",
        attrib=attributes,
        )
    if transform:
        group.set("transform", transform)

    if title is not None:
        xml.SubElement(group, "title").text = str(title)

    fonts = toyplot.font.ReportlabLibrary()
    layout = text if isinstance(text, toyplot.text.Layout) else toyplot.text.layout(text, style, fonts)
    if layout.style.get("-toyplot-text-layout-visibility", None) == "visible":
        xml.SubElement(
            group,
            "rect",
            x=str(layout.left),
            y=str(layout.top),
            width=str(layout.width),
            height=str(layout.height),
            stroke="red",
            fill="none",
            opacity="0.5",
            )
        xml.SubElement(
            group,
            "circle",
            x="0",
            y="0",
            r="3",
            stroke="none",
            fill="red",
            opacity="0.5",
            )

    hyperlink = []
    for line in layout.children:
        if line.style.get("-toyplot-text-layout-line-visibility", None) == "visible":
            xml.SubElement(
                group,
                "rect",
                x=str(line.left),
                y=str(line.top),
                width=str(line.width),
                height=str(line.height),
                stroke="green",
                fill="none",
                opacity="0.5",
                )
            xml.SubElement(
                group,
                "line",
                x1=str(line.left),
                y1=str(line.baseline),
                x2=str(line.right),
                y2=str(line.baseline),
                stroke="green",
                fill="none",
                opacity="0.5",
                )
        for box in line.children:
            if isinstance(box, toyplot.text.TextBox):
                xml.SubElement(
                    group,
                    "text",
                    x=str(box.left),
                    y=str(box.baseline),
                    style=toyplot.style.to_css(style), #toyplot.style.to_css(box.style),
                    ).text = box.text

                if box.style.get("-toyplot-text-layout-box-visibility", None) == "visible":
                    xml.SubElement(
                        group,
                        "rect",
                        x=str(box.left),
                        y=str(box.top),
                        width=str(box.width),
                        height=str(box.height),
                        stroke="blue",
                        fill="none",
                        opacity="0.5",
                        )
                    xml.SubElement(
                        group,
                        "line",
                        x1=str(box.left),
                        y1=str(box.baseline),
                        x2=str(box.right),
                        y2=str(box.baseline),
                        stroke="blue",
                        fill="none",
                        opacity="0.5",
                        )

            elif isinstance(box, toyplot.text.MarkerBox):
                if box.marker:
                    _draw_marker(
                        group,
                        cx=(box.left + box.right) * 0.5,
                        cy=(box.top + box.bottom) * 0.5,
                        marker=toyplot.marker.create(size=box.height) + box.marker,
                        )
                if box.style.get("-toyplot-text-layout-box-visibility", None) == "visible":
                    xml.SubElement(
                        group,
                        "rect",
                        x=str(box.left),
                        y=str(box.top),
                        width=str(box.width),
                        height=str(box.height),
                        stroke="blue",
                        fill="none",
                        opacity="0.5",
                        )
                    xml.SubElement(
                        group,
                        "line",
                        x1=str(box.left),
                        y1=str(box.baseline),
                        x2=str(box.right),
                        y2=str(box.baseline),
                        stroke="blue",
                        fill="none",
                        opacity="0.5",
                        )

            elif isinstance(box, toyplot.text.PushHyperlink):
                hyperlink.append(group)
                group = xml.SubElement(group, "a")
                group.set("xlink:href", box.href)
            elif isinstance(box, toyplot.text.PopHyperlink):
                group = hyperlink.pop()



## copy from this to draw my own graph renderer
#@dispatch(toyplot.coordinates.Cartesian, toyplot.mark.Graph, RenderContext)
def _render_graph(axes, mark, context): 
    # Project edge coordinates.
    for i in range(2):
        if mark._coordinate_axes[i] == "x":
            edge_x = axes.project("x", mark._ecoordinates.T[i])
        elif mark._coordinate_axes[i] == "y":
            edge_y = axes.project("y", mark._ecoordinates.T[i])
    edge_coordinates = numpy.column_stack((edge_x, edge_y))

    # Project vertex coordinates.
    for i in range(2):
        if mark._coordinate_axes[i] == "x":
            vertex_x = axes.project("x", mark._vtable[mark._vcoordinates[i]])
        elif mark._coordinate_axes[i] == "y":
            vertex_y = axes.project("y", mark._vtable[mark._vcoordinates[i]])

    # NO VERTEX LABELS
    #vertex_markers = []
    #for vmarker, vsize, vcolor, vopacity in zip(
    #        mark._vtable[mark._vmarker[0]],
    #        mark._vtable[mark._vsize[0]],
    #        mark._vtable[mark._vcolor[0]],
    #        mark._vtable[mark._vopacity[0]],
    #    ):
    #    if vmarker:
    #        vstyle = toyplot.style.combine(
    #            {
    #            "fill": toyplot.color.to_css(vcolor),
    #            "stroke": toyplot.color.to_css(vcolor),
    #            "opacity": vopacity,
    #            },
    #            mark._vstyle)
    #        vertex_marker = toyplot.marker.create(size=vsize, mstyle=vstyle, lstyle=mark._vlstyle) + toyplot.marker.convert(vmarker)
    #        vertex_markers.append(vertex_marker)
    #    else:
    #        vertex_markers.append(None)

    # EDGE STYLES
    edge_styles = []
    for ecolor, ewidth, eopacity in zip(
            mark._etable[mark._ecolor[0]],
            mark._etable[mark._ewidth[0]],
            mark._etable[mark._eopacity[0]],
        ):
        edge_styles.append(
            toyplot.style.combine(
                {
                "fill": "none",
                "stroke": toyplot.color.to_css(ecolor),
                "stroke-width": ewidth,
                "stroke-opacity": eopacity,
                },
                mark._estyle,
            )
        )

    edge_marker_styles = []
    for ecolor, estyle in zip(
            mark._etable[mark._ecolor[0]],
            edge_styles,
            ):
        edge_marker_styles.append(toyplot.style.combine(estyle, {"fill": toyplot.color.to_css(ecolor)}))

    # Identify ranges of edge coordinates for each edge.
    index = 0
    edge_start = []
    edge_end = []
    for eshape in mark._etable[mark._eshape[0]]:
        edge_start.append(index)
        for segment in eshape:
            if segment == "M":
                count = 1
            elif segment == "L":
                count = 1
            elif segment == "Q":
                count = 2
            elif segment == "C":
                count = 3
            index += count
        edge_end.append(index)

    # Adjust edge coordinates so edges don't overlap vertex markers.
    for esource, etarget, start, end in zip(
            mark._etable[mark._esource[0]],
            mark._etable[mark._etarget[0]],
            edge_start,
            edge_end,
        ):

        # Skip loop edges.
        if esource == etarget:
            continue

        source_vertex_marker = vertex_markers[esource]
        target_vertex_marker = vertex_markers[etarget]

        if source_vertex_marker:
            dp = source_vertex_marker.intersect(edge_coordinates[start + 1] - edge_coordinates[start])
            edge_coordinates[start] += dp

        if target_vertex_marker:
            dp = target_vertex_marker.intersect(edge_coordinates[end - 2] - edge_coordinates[end - 1])
            edge_coordinates[end - 1] += dp

    # Render the graph.
    mark_xml = xml.SubElement(
        context.parent, "g", 
        id=context.get_id(mark), 
        attrib={"class": "toyplot-mark-Graph"}
        )
    #_render_table(
    #    owner=mark, 
    #    key="vertex_data", 
    #    label="graph vertex data", 
    #    table=mark._vtable, 
    #    filename=mark._vfilename, 
    #    context=context)

    ## THIS ADDS THE DOWNLOAD INFO (NEWICK STRING FILE) TO THE GRAPH
    _render_table(
        owner=mark, 
        key="edge_data", 
        label="graph edge data", 
        table=mark._etable, 
        filename=mark._efilename, 
        context=context
        )

    # Render edges.
    ## THIS IS WHERE WE STICK IN THE COMMON STYLE ELEMENTS
    edge_xml = xml.SubElement(
        mark_xml, 
        "g", 
        attrib={"class": "toyplot-Edges"}
        )
    ## THEN PUT IN THE UNIQUE STYLE ELEMENTS
    for esource, etarget, eshape, estyle, hmarker, mmarker, mposition, tmarker, start, end in zip(
            mark._etable[mark._esource[0]],
            mark._etable[mark._etarget[0]],
            mark._etable[mark._eshape[0]],
            edge_styles,
            mark._etable[mark._hmarker[0]],
            mark._etable[mark._mmarker[0]],
            mark._etable[mark._mposition[0]],
            mark._etable[mark._tmarker[0]],
            edge_start,
            edge_end,
        ):

        path = []
        index = 0
        for segment in eshape:
            if segment == "M":
                count = 1
            elif segment == "L":
                count = 1
            elif segment == "Q":
                count = 2
            elif segment == "C":
                count = 3
            path.append(segment)
            for _ in range(count):
                path.append(str(edge_coordinates[start + index][0]))
                path.append(str(edge_coordinates[start + index][1]))
                index += 1

        xml.SubElement(
            edge_xml,
            "path",
            d=" ".join(path),
            style=_css_style(estyle),
            )

    # Render edge head markers.
    # marker_xml = xml.SubElement(edge_xml, "g", attrib={"class": "toyplot-HeadMarkers"})
    # for marker, mstyle, estart, eend in zip(
    #         mark._etable[mark._hmarker[0]],
    #         edge_marker_styles,
    #         edge_start,
    #         edge_end,
    #     ):
    #     if marker:
    #         # Create the marker with defaults.
    #         marker = toyplot.marker.create(size=10, mstyle=mstyle) + toyplot.marker.convert(marker)

    #         # Compute the marker angle using the first edge segment.
    #         edge_angle = -numpy.rad2deg(numpy.arctan2(
    #             edge_coordinates[estart+1][1] - edge_coordinates[estart][1],
    #             edge_coordinates[estart+1][0] - edge_coordinates[estart][0],
    #             ))

    #         transform = "translate(%r, %r)" % (edge_coordinates[estart][0], edge_coordinates[estart][1])
    #         if edge_angle:
    #             transform += " rotate(%r)" % (-edge_angle,)
    #         transform += " translate(%r, 0)" % (marker.size / 2,)
    #         if marker.angle is not None:
    #             if isinstance(marker.angle, toyplot.compatibility.string_type) and marker.angle[0:1] == "r":
    #                 angle = float(marker.angle[1:])
    #             else:
    #                 angle = -edge_angle + float(marker.angle)
    #             transform += " rotate(%r)" % (-angle,)

    #         _draw_marker(
    #             marker_xml,
    #             marker=marker,
    #             transform=transform,
    #             )

    # # Render edge middle markers.
    # marker_xml = xml.SubElement(edge_xml, "g", attrib={"class": "toyplot-MiddleMarkers"})
    # for mstyle, marker, mposition, start, end in zip(
    #         edge_marker_styles,
    #         mark._etable[mark._mmarker[0]],
    #         mark._etable[mark._mposition[0]],
    #         edge_start,
    #         edge_end,
    #     ):
    #     if marker:
    #         # Create the marker with defaults.
    #         marker = toyplot.marker.create(size=10, mstyle=mstyle) + toyplot.marker.convert(marker)

    #         # Place the marker within the first edge segment.
    #         x, y = edge_coordinates[start] * (1 - mposition) + edge_coordinates[start+1] * mposition

    #         # Compute the marker angle using the first edge segment.
    #         angle = -numpy.rad2deg(numpy.arctan2(
    #             edge_coordinates[start+1][1] - edge_coordinates[start][1],
    #             edge_coordinates[start+1][0] - edge_coordinates[start][0],
    #             ))
    #         if marker.angle is not None:
    #             if isinstance(marker.angle, toyplot.compatibility.string_type) and marker.angle[0:1] == "r":
    #                 angle += float(marker.angle[1:])
    #             else:
    #                 angle = float(marker.angle)

    #         marker = marker + toyplot.marker.create(angle=angle)

    #         _draw_marker(
    #             marker_xml,
    #             cx=x,
    #             cy=y,
    #             marker=marker,
    #             )

    # # Render edge tail markers.
    # marker_xml = xml.SubElement(edge_xml, "g", attrib={"class": "toyplot-TailMarkers"})
    # for mstyle, marker, start, end in zip(
    #         edge_marker_styles,
    #         mark._etable[mark._tmarker[0]],
    #         edge_start,
    #         edge_end,
    #     ):
    #     if marker:
    #         # Create the marker with defaults.
    #         marker = toyplot.marker.create(size=10, mstyle=mstyle, lstyle={}) + toyplot.marker.convert(marker)

    #         # Compute the marker angle using the last edge segment.
    #         edge_angle = -numpy.rad2deg(numpy.arctan2(
    #             edge_coordinates[end-1][1] - edge_coordinates[end-2][1],
    #             edge_coordinates[end-1][0] - edge_coordinates[end-2][0],
    #             ))

    #         transform = "translate(%r, %r)" % (edge_coordinates[end-1][0], edge_coordinates[end-1][1])
    #         if edge_angle:
    #             transform += " rotate(%r)" % (-edge_angle,)
    #         transform += " translate(%r, 0)" % (-marker.size / 2,)
    #         if marker.angle is not None:
    #             if isinstance(marker.angle, toyplot.compatibility.string_type) and marker.angle[0:1] == "r":
    #                 angle = float(marker.angle[1:])
    #             else:
    #                 angle = -edge_angle + float(marker.angle)
    #             transform += " rotate(%r)" % (-angle,)


    #         _draw_marker(
    #             marker_xml,
    #             marker=marker,
    #             transform=transform,
    #             )

    # Render vertex markers
    #vertex_xml = xml.SubElement(mark_xml, "g", attrib={"class": "toyplot-Vertices"})
    #for vx, vy, vmarker, vtitle in zip(
    #        vertex_x,
    #        vertex_y,
    #        vertex_markers,
    #        mark._vtable[mark._vtitle[0]],
    #    ):
    #    if vmarker:
    #        _draw_marker(
    #            vertex_xml,
    #            cx=vx,
    #            cy=vy,
    #            marker=vmarker,
    #            extra_class="toyplot-Datum",
    #            title=vtitle,
    #            )

    # Render vertex labels
    #if mark._vlshow:
    #    vlabel_xml = xml.SubElement(mark_xml, "g", attrib={"class": "toyplot-Labels"})
    #    for dx, dy, dtext in zip(vertex_x, vertex_y, mark._vtable[mark._vlabel[0]]):
    #        _draw_text(
    #            root=vlabel_xml,
    #            text=toyplot.compatibility.unicode_type(dtext),
    #            x=dx,
    #            y=dy,
    #            style=mark._vlstyle,
    #            attributes={"class": "toyplot-Datum"},
    #            )

